"""CLI for running commands in parallel"""

import contextlib
import enum
from collections import OrderedDict
from pathlib import Path
from typing import Annotated, Any, Optional

import anyio
import rich
import typer

from .executor import Command, CommandGroup, CommandStatus, ProcessingStrategy, read_commands_toml

PID_FILE = ".par-run.uvicorn.pid"

cli_app = typer.Typer()


# Web only functions
def clean_up() -> None:
    """Clean up by removing the PID file."""
    Path(PID_FILE).unlink()
    typer.echo("Cleaned up PID file.")


def start_web_server(port: int) -> None:
    """Start the web server"""
    if Path(PID_FILE).is_file():
        typer.echo("UVicorn server is already running.")
        sys.exit(1)
    with Path(PID_FILE).open("w", encoding="utf-8") as pid_file:
        typer.echo(f"Starting UVicorn server on port {port}...")
        uvicorn_command = [
            "uvicorn",
            "par_run.web:ws_app",
            "--host",
            "0.0.0.0",
            "--port",
            str(port),
        ]
        process = subprocess.Popen(uvicorn_command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        pid_file.write(str(process.pid))

        # Wait for UVicorn to start
        wait_time = 3 * 10**9  # 3 seconds
        start_time = time.time_ns()

        while time.time_ns() - start_time < wait_time:
            test_port = get_process_port(process.pid)
            if port == test_port:
                typer.echo(f"UVicorn server is running on port {port} in {(time.time_ns() - start_time)/10**6:.2f} ms.")
                typer.echo(f"Server running at http://localhost:{port}/")
                break
            time.sleep(0.1)  # Poll every 0.1 seconds

        else:
            typer.echo(f"UVicorn server did not respond within {wait_time} seconds.")
            typer.echo("run 'par-run web status' to check the status.")


def stop_web_server() -> None:
    """Stop the UVicorn server by reading its PID from the PID file and sending a termination signal."""
    if not Path(PID_FILE).is_file():
        typer.echo("UVicorn server is not running.")
        return

    with Path(PID_FILE).open() as pid_file:
        pid = int(pid_file.read().strip())

    typer.echo(f"Stopping UVicorn server with {pid=:}...")
    with contextlib.suppress(ProcessLookupError):
        os.kill(pid, signal.SIGTERM)
    clean_up()


def get_process_port(pid: int) -> Optional[int]:
    process = psutil.Process(pid)
    connections = process.connections()
    if connections:
        port = connections[0].laddr.port
        return port
    return None


def list_uvicorn_processes() -> None:
    """Check for other UVicorn processes and list them"""
    uvicorn_processes = []
    with contextlib.suppress(psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
        for process in psutil.process_iter():
            process_name = process.name()
            if "uvicorn" in process_name.lower():
                uvicorn_processes.append(process)

    if uvicorn_processes:
        typer.echo("Other UVicorn processes:")
        for process in uvicorn_processes:
            typer.echo(f"PID: {process.pid}, Name: {process.name()}")
    else:
        typer.echo("No other UVicorn processes found.")


def get_web_server_status() -> None:
    """Get the status of the UVicorn server by reading its PID from the PID file."""
    if not Path(PID_FILE).is_file():
        typer.echo("No pid file found. Server likely not running.")
        list_uvicorn_processes()
        return

    with Path(PID_FILE).open() as pid_file:
        pid = int(pid_file.read().strip())
        if psutil.pid_exists(pid):
            port = get_process_port(pid)
            if port:
                typer.echo(f"UVicorn server is running with {pid=}, {port=}")
            else:
                typer.echo(f"UVicorn server is running with {pid=:}, couldn't determine port.")
        else:
            typer.echo("UVicorn server is not running but pid files exists, deleting it.")
            clean_up()


class WebCommand(enum.Enum):
    """Web command enumeration."""

    START = "start"
    STOP = "stop"
    RESTART = "restart"
    STATUS = "status"

    def __str__(self) -> str:
        return self.value


class AsyncBackend(enum.Enum):
    """Async backend enumeration."""

    TRIO = "trio"
    ASYNCIO = "asyncio"
    ASYNCIO_NATIVE = "asyncio-native"

    def __str__(self) -> str:
        return self.value


class CLICommandCBOnComp:
    async def on_start(self, cmd: Command) -> None:
        rich.print(f"[blue bold]{cmd.name}: Started[/]")

    async def on_recv(self, _: Command, output: str) -> None:
        rich.print(output)

    async def on_term(self, cmd: Command, exit_code: int) -> None:
        """Callback function for when a command receives output"""
        if cmd.status == CommandStatus.SUCCESS:
            rich.print(f"[green bold]{cmd.name}: Finished[/]")
        elif cmd.status == CommandStatus.FAILURE:
            rich.print(f"[red bold]{cmd.name}: Failed, {exit_code=:}[/]")


class CLICommandCBOnRecv:
    async def on_start(self, cmd: Command) -> None:
        rich.print(f"[blue bold]{cmd.name}: Started[/]")

    async def on_recv(self, cmd: Command, output: str) -> None:
        rich.print(f"{cmd.name}: {output}")

    async def on_term(self, cmd: Command, exit_code: int) -> None:
        """Callback function for when a command receives output"""
        if cmd.status == CommandStatus.SUCCESS:
            rich.print(f"[green bold]{cmd.name}: Finished[/]")
        elif cmd.status == CommandStatus.FAILURE:
            rich.print(f"[red bold]{cmd.name}: Failed, {exit_code=:}[/]")


def format_elapsed_time(seconds: float) -> str:
    """Converts a number of seconds into a human-readable time format of HH:MM:SS.xxx

    Args:
    ----
    seconds (float): The number of seconds elapsed.

    Returns:
    -------
    str: The formatted time string.

    """
    hours = int(seconds) // 3600
    minutes = (int(seconds) % 3600) // 60
    seconds = seconds % 60  # Keeping the fractional part of seconds

    # Return formatted string with seconds rounded to 2 d.p.
    return f"{hours:02}:{minutes:02}:{seconds:06.3f}"


def show_commands(groups: list[CommandGroup]) -> None:
    for grp in groups:
        rich.print(f"[blue bold]Group: {grp.name}[/]")
        rich.print(f"Params: cont_on_fail={grp.cont_on_fail}, serial={grp.serial}, timeout={grp.timeout}")
        for cmd in grp.cmds.values():
            rich.print(f"[green bold]{cmd.name}[/]: {cmd.cmd}")


def filter_groups(
    group_list: list[CommandGroup], filter_groups: Optional[str], filter_cmds: Optional[str]
) -> list[CommandGroup]:
    if filter_groups:
        group_list = [grp for grp in group_list if grp.name in [g.strip() for g in filter_groups.split(",")]]

    if filter_cmds:
        for grp in group_list:
            grp.cmds = OrderedDict(
                {
                    cmd_name: cmd
                    for cmd_name, cmd in grp.cmds.items()
                    if cmd_name in [c.strip() for c in filter_cmds.split(",")]
                },
            )
        group_list = [grp for grp in group_list if grp.cmds]
    return group_list


def add_table_break(tbl: rich.table.Table, break_ch: str = "-", break_style: Optional[str] = None) -> rich.table.Table:
    break_data: list[str] = [break_ch * int(col.width) for col in tbl.columns if col.width is not None]
    tbl.add_row(
        *break_data,
        style=break_style,
    )
    return tbl


def build_results_tbl() -> rich.table.Table:
    res_tbl = rich.table.Table(title="Results", show_header=True, header_style="bold blue", box=rich.box.ROUNDED)
    group_w, name_w, cmd_w, status_w, elap_w = (10, 15, 50, 6, 12)
    res_tbl.add_column("Group", style="bold blue", width=group_w, no_wrap=True)
    res_tbl.add_column("Name", style="bold blue", width=name_w, no_wrap=True)
    res_tbl.add_column("Command", style="bold blue", width=cmd_w, no_wrap=True)
    res_tbl.add_column("Status", style="bold blue", width=status_w, no_wrap=True)
    res_tbl.add_column("Elapsed", style="bold blue", width=elap_w, no_wrap=True)
    return res_tbl


def command_status_to_emoji(cmd: Command) -> tuple[str, str]:
    if cmd.status == CommandStatus.SUCCESS:
        return ("✅", "green")
    elif cmd.status == CommandStatus.FAILURE:
        return ("❌", "red")
    elif cmd.status == CommandStatus.TIMEOUT:
        return ("⏰", "orange1")
    else:
        return ("🚫", "red")


def add_command_row(tbl: rich.table.Table, cmd: Command, group_name: str) -> rich.table.Table:
    elap_str = format_elapsed_time(cmd.elapsed) if cmd.elapsed else "XX:XX:XX.xxx"
    emoji, row_style = command_status_to_emoji(cmd)
    tbl.add_row(group_name, cmd.name, cmd.cmd, emoji, elap_str, style=row_style)
    return tbl


def fmt_group_name(cmd_group: CommandGroup) -> str:
    if cmd_group.status == CommandStatus.SUCCESS:
        return f"[green]{cmd_group.name}[/]"
    elif cmd_group.status == CommandStatus.FAILURE:
        return f"[red]{cmd_group.name}[/]"
    else:
        return f"[yellow]{cmd_group.name}[/]"


style_default = typer.Option(help="Processing strategy, for serial commands this is always --recv", default="comp")
show_default = typer.Option(help="Show available groups and commands", default=False)
pyproj_default = typer.Option(help="The default toml file to use", default=Path("pyproject.toml"))
groups_default = typer.Option(help="Run a specific group of commands, comma spearated", default=None)
cmds_default = typer.Option(help="Run specific commands, comma separated", default=None)
backend_default = typer.Option(help="The backend to use", default="trio")

backend_options: dict[AsyncBackend, Any] = {
    AsyncBackend.TRIO: {"backend": "trio", "backend_options": {}},
    AsyncBackend.ASYNCIO: {"backend": "asyncio", "backend_options": {"use_uvloop": True}},
    AsyncBackend.ASYNCIO_NATIVE: {"backend": "asyncio", "backend_options": {"use_uvloop": False}},
}


@cli_app.command()
def run(  # noqa: PLR0913
    style: Annotated[ProcessingStrategy, typer.Option] = style_default,
    show: Annotated[bool, typer.Option] = show_default,
    file: Annotated[Path, typer.Option] = pyproj_default,
    groups: Annotated[Optional[str], typer.Option] = groups_default,
    cmds: Annotated[Optional[str], typer.Option] = cmds_default,
    backend: Annotated[AsyncBackend, typer.Option] = backend_default,
) -> None:
    """Run commands in parallel"""
    # Overall exit code, need to track all command exit codes to update this
    exit_code = 0
    st_all = time.perf_counter()

    actual_backend = backend_options[backend]["backend"]
    actual_backend_opts = backend_options[backend]["backend_options"]

    master_groups = read_commands_toml(file)
    if show:
        return show_commands(master_groups)

    master_groups = filter_groups(master_groups, groups, cmds)

    if not master_groups:
        rich.print("[blue]No groups or commands found.[/]")
        raise typer.Exit(0)

    for grp in master_groups:
        # Run the async function with Trio's event loop
        if style == ProcessingStrategy.ON_COMP:
            anyio.run(grp.run, style, CLICommandCBOnComp(), backend=actual_backend, backend_options=actual_backend_opts)
        elif style == ProcessingStrategy.ON_RECV:
            anyio.run(grp.run, style, CLICommandCBOnRecv(), backend=actual_backend, backend_options=actual_backend_opts)
        else:
            raise typer.BadParameter("Invalid processing strategy")  # pragma: no cover
        if grp.status != CommandStatus.SUCCESS:
            exit_code = 1
            if not grp.cont_on_fail:
                break
    # Summarise the results
    console = rich.console.Console()
    res_tbl = build_results_tbl()

    for grp_ix, grp in enumerate(master_groups):
        for ix, cmd in enumerate(grp.cmds.values()):
            if grp_ix > 0 and ix == 0:
                add_table_break(res_tbl)
            grp_name = fmt_group_name(grp)
            if ix > 0:
                grp_name = ""
            add_command_row(res_tbl, cmd, grp_name)

    console.print(res_tbl)
    end_style = "[green bold]" if exit_code == 0 else "[red bold]"
    rich.print(f"\n{end_style}Total elapsed time: {format_elapsed_time(time.perf_counter() - st_all)}[/]")
    raise typer.Exit(exit_code)


try:
    import os
    import signal
    import subprocess
    import sys
    import time
    from pathlib import Path
    from typing import Optional

    import psutil
    import typer

    rich.print("[blue]Web commands loaded[/]")

    PID_FILE = ".par-run.uvicorn.pid"

    command_default = typer.Argument(..., help="command to control/interract with the web server")
    port_default = typer.Option(8001, help="Port to run the web server")

    @cli_app.command()
    def web(
        command: WebCommand = command_default,
        port: int = port_default,
    ) -> None:
        """Run the web server"""
        if command == WebCommand.START:
            start_web_server(port)
        elif command == WebCommand.STOP:
            stop_web_server()
        elif command == WebCommand.RESTART:
            stop_web_server()
            start_web_server(port)
        elif command == WebCommand.STATUS:
            get_web_server_status()
        else:
            typer.echo(f"Not a valid command '{command}'", err=True)  # pragma: no cover
            raise typer.Abort()  # pragma: no cover

except ImportError:  # pragma: no cover
    pass  # pragma: no cover



"""Todo"""

import enum
import os
import subprocess
import time
from collections import OrderedDict
from collections.abc import AsyncIterable
from pathlib import Path
from typing import Any, Optional, Protocol, Union, cast

import anyio
import anyio.to_thread
import tomlkit
from pydantic import BaseModel, ConfigDict, Field

internal_error_ret_code = 999


class ProcessingStrategy(enum.Enum):
    """Enum for processing strategies."""

    ON_COMP = "comp"
    ON_RECV = "recv"


class CommandStatus(enum.Enum):
    """Enum for command status."""

    NOT_STARTED = "Not Started"
    RUNNING = "Running"
    SUCCESS = "Success"
    FAILURE = "Failure"
    TIMEOUT = "Timeout"
    CANCELLED = "Cancelled"

    def completed(self) -> bool:
        """Return True if the command has completed."""
        return self in [CommandStatus.SUCCESS, CommandStatus.FAILURE, CommandStatus.TIMEOUT, CommandStatus.CANCELLED]


class Command(BaseModel):
    """Holder for a command and its name."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    name: str
    cmd: str
    passenv: Optional[list[str]] = Field(default=None)
    setenv: Optional[dict[str, str]] = Field(default=None)
    status: CommandStatus = CommandStatus.NOT_STARTED
    unflushed: list[str] = Field(default=[], exclude=True)
    num_non_empty_lines: int = Field(default=0, exclude=True)
    ret_code: Optional[int] = Field(default=None, exclude=True)
    start_time: Optional[float] = Field(default=None, exclude=True)
    elapsed: Optional[float] = Field(default=None, exclude=True)

    def incr_line_count(self, line: str) -> None:
        """Increment the non-empty line count."""
        if line.strip():
            self.num_non_empty_lines += 1

    def append_unflushed(self, line: str) -> None:
        """Append a line to the output and increment the non-empty line count."""
        self.unflushed.append(line)

    def clear_unflushed(self) -> None:
        """Clear the unflushed output."""
        self.unflushed.clear()

    def set_ret_code(self, ret_code: int) -> None:
        """Set the return code and status of the command."""
        if self.start_time:
            self.elapsed = time.perf_counter() - self.start_time
        self.ret_code = ret_code
        if ret_code == 0:
            self.status = CommandStatus.SUCCESS
        else:
            self.status = CommandStatus.FAILURE

    def set_timeout(self) -> None:
        """Set the command status to timeout."""
        self.status = CommandStatus.TIMEOUT
        if self.start_time:
            self.elapsed = time.perf_counter() - self.start_time
        self.ret_code = internal_error_ret_code

    def set_running(self) -> None:
        """Set the command status to running."""
        self.start_time = time.perf_counter()
        self.status = CommandStatus.RUNNING

    def set_cancelled(self) -> None:
        """Set the command status to cancelled."""
        self.status = CommandStatus.CANCELLED
        self.ret_code = internal_error_ret_code


class CommandCB(Protocol):
    async def on_start(self, cmd: Command) -> None: ...
    async def on_recv(self, cmd: Command, output: str) -> None: ...
    async def on_term(self, cmd: Command, exit_code: int) -> None: ...


class CommandGroup(BaseModel):
    """Holder for a group of commands."""

    name: str
    desc: Optional[str] = None
    cmds: OrderedDict[str, Command] = Field(default_factory=OrderedDict)
    timeout: int = Field(default=30)
    cont_on_fail: bool = Field(default=False)
    serial: bool = Field(default=False)
    status: CommandStatus = CommandStatus.NOT_STARTED

    def update_status(self, cmds: OrderedDict[str, Command]) -> None:
        """Update the status of the command group."""
        if all(cmd.status == CommandStatus.SUCCESS for cmd in cmds.values()):
            self.status = CommandStatus.SUCCESS
        else:
            self.status = CommandStatus.FAILURE

    async def run_serial(self, callbacks: CommandCB) -> None:
        try:
            for ix in range(len(self.cmds.values())):
                cmd = list(self.cmds.values())[ix]
                cmd.set_running()
                async with anyio.create_task_group() as nursery:
                    nursery.start_soon(self._run_command, cmd, ProcessingStrategy.ON_RECV, callbacks)
                if cmd.status != CommandStatus.SUCCESS and not self.cont_on_fail:
                    # Cancel all remaining cmds
                    for jx in range(ix + 1, len(self.cmds.values())):
                        cmd = list(self.cmds.values())[jx]
                        cmd.set_cancelled()
                    self.update_status(self.cmds)
                    break
        except Exception as _:
            self.status = CommandStatus.FAILURE
        else:
            self.update_status(self.cmds)

    async def run_parallel(self, strategy: ProcessingStrategy, callbacks: CommandCB) -> None:
        try:
            async with anyio.create_task_group() as nursery:
                for cmd in self.cmds.values():
                    nursery.start_soon(self._run_command, cmd, strategy, callbacks)
                    cmd.set_running()
        except Exception as _:
            self.status = CommandStatus.FAILURE
        else:
            self.update_status(self.cmds)

    async def run(self, strategy: ProcessingStrategy, callbacks: CommandCB) -> None:
        if self.serial:
            await self.run_serial(callbacks)
        else:
            await self.run_parallel(strategy, callbacks)

    async def _proces_stdxx_line(
        self, cmd: Command, line: str, strategy: ProcessingStrategy, callbacks: CommandCB
    ) -> None:
        if strategy == ProcessingStrategy.ON_RECV:
            await callbacks.on_recv(cmd, line)
        elif strategy == ProcessingStrategy.ON_COMP:
            cmd.append_unflushed(line)
        cmd.incr_line_count(line)

    async def _process_stdxxx(
        self, cmd: Command, strategy: ProcessingStrategy, stream: AsyncIterable[bytes], callbacks: CommandCB
    ) -> None:
        await callbacks.on_start(cmd)

        incomplete_line = ""

        async for chunk in stream:
            # Decode the bytes to a string
            decoded_chunk = chunk.decode("utf-8")
            # Combine the remainder of the last chunk with the new chunk
            lines = (incomplete_line + decoded_chunk).split("\n")

            # The last line might be incomplete; hold it back
            incomplete_line = lines.pop() if lines[-1] else ""

            for line in lines:
                await self._proces_stdxx_line(cmd, line, strategy, callbacks)

        if incomplete_line:
            await self._proces_stdxx_line(cmd, incomplete_line, strategy, callbacks)  # pragma: no cover

        if strategy == ProcessingStrategy.ON_COMP:
            for _ix, line in enumerate(cmd.unflushed):
                await callbacks.on_recv(cmd, line)
            cmd.clear_unflushed()

    async def _run_command(self, cmd: Command, strategy: ProcessingStrategy, callbacks: CommandCB) -> int:
        env = os.environ.copy()
        if cmd.setenv:
            env.update(cmd.setenv)

        # Running the command asynchronously and capturing the output
        try:
            with anyio.fail_after(self.timeout):
                async with (
                    await anyio.open_process(
                        command=cmd.cmd, env=env, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
                    ) as process,
                    anyio.create_task_group() as tg,
                ):
                    if process.stdout:
                        tg.start_soon(self._process_stdxxx, cmd, strategy, process.stdout, callbacks)
                    await process.wait()
        except TimeoutError:
            cmd.set_timeout()
            await callbacks.on_term(cmd, internal_error_ret_code)
            return internal_error_ret_code
        else:
            exit_code = cast(int, process.returncode)
            cmd.set_ret_code(exit_code)
            await callbacks.on_term(cmd, exit_code)
            return exit_code


def _validate_mandatory_keys(data: tomlkit.items.Table, keys: list[str], context: str) -> tuple[Any, ...]:
    """Validate that the mandatory keys are present in the data.

    Args:
    ----
        data (tomlkit.items.Table): The data to validate.
        keys (list[str]): The mandatory keys.

    """
    vals = []
    for key in keys:
        val = data.get(key, None)
        if not val:
            raise ValueError(f"{key} is mandatory, not found in {context}")
        vals.append(val)
    return tuple(vals)


def _get_optional_keys(data: tomlkit.items.Table, keys: list[str], default: Any = None) -> tuple[Optional[Any], ...]:  # noqa: ANN001, ANN401
    """Get Optional keys or default.

    Args:
    ----
        data (tomlkit.items.Table): The data to use as source
        keys (list[str]): The optional keys.

    """
    res = tuple(data.get(key, default) for key in keys)
    return res


def read_commands_toml(filename: Union[str, Path]) -> list[CommandGroup]:
    """Read a commands.toml file and return a list of CommandGroup objects.

    Args:
    ----
        filename (Union[str, Path]): The filename of the commands.toml file.

    Returns:
    -------
        list[CommandGroup]: A list of CommandGroup objects.

    """
    with Path(filename).open(encoding="utf-8") as toml_file:
        toml_data = tomlkit.parse(toml_file.read())

    if (isinstance(filename, Path) and filename.name == "pyproject.toml") or (
        isinstance(filename, str) and filename == "pyproject.toml"
    ):
        cmd_groups_data = toml_data.get("tool", {}).get("par-run", {})
    else:
        cmd_groups_data = toml_data.get("par-run", None)

    if not cmd_groups_data:
        raise ValueError(f"No par-run data found in toml file {filename}")
    _ = cmd_groups_data.get("description", None)

    command_groups = []
    for group_data in cmd_groups_data.get("groups", []):
        (group_name,) = _validate_mandatory_keys(group_data, ["name"], "top level par-run group")
        group_desc, group_timeout = _get_optional_keys(
            group_data,
            ["desc", "timeout"],
            default=None,
        )
        (group_cont_on_fail, group_serial) = _get_optional_keys(group_data, ["cont_on_fail", "serial"], default=False)

        if not group_timeout:
            group_timeout = 30
        group_cont_on_fail = bool(group_cont_on_fail and group_cont_on_fail is True)
        group_serial = bool(group_serial and group_serial is True)

        commands = OrderedDict()
        for cmd_data in group_data.get("commands", []):
            name, exec = _validate_mandatory_keys(cmd_data, ["name", "exec"], f"command group {group_name}")
            setenv, passenv = _get_optional_keys(cmd_data, ["setenv", "passenv"], default=None)

            commands[name] = Command(name=name, cmd=exec, setenv=setenv, passenv=passenv)
        command_group = CommandGroup(
            name=group_name,
            desc=group_desc,
            cmds=commands,
            timeout=group_timeout,
            cont_on_fail=group_cont_on_fail,
            serial=group_serial,
        )
        command_groups.append(command_group)

    return command_groups



from collections import OrderedDict
from collections.abc import Generator

import pytest
import requests
from fastapi.testclient import TestClient
from httpx import AsyncClient
from pytest_mock import MockerFixture

from par_run.executor import Command, CommandGroup
from par_run.web import ws_app

from .conftest import AnyIOBackendT


@pytest.fixture(
    params=[
        pytest.param(("asyncio", {"use_uvloop": True}), id="asyncio+uvloop"),
        pytest.param(("asyncio", {"use_uvloop": False}), id="asyncio"),
    ]
)
def anyio_backend(request: pytest.FixtureRequest) -> AnyIOBackendT:
    return request.param  # type: ignore


@pytest.fixture()
def test_client(anyio_backend: AnyIOBackendT) -> Generator[TestClient, None, None]:
    yield TestClient(ws_app, backend=anyio_backend[0], backend_options=anyio_backend[1])


@pytest.fixture()
def async_client() -> Generator[AsyncClient, None, None]:
    yield AsyncClient(app=ws_app, base_url="http://test")


def test_ws_main() -> None:
    client = TestClient(ws_app)
    response = client.get("/")
    http_ok = 200
    assert response.status_code == http_ok


def test_ws_full(test_client: TestClient, mocker: MockerFixture, anyio_backend: AnyIOBackendT) -> None:  # noqa: ARG001
    command1 = Command(name="test1", cmd="echo 'Hello, World!'")
    command2 = Command(name="test2", cmd="echo 'World, Hey!'")
    commands = OrderedDict()
    commands[command1.name] = command1
    commands[command2.name] = command2
    groups = [CommandGroup(name="test_group", cmds=commands)]

    mocker.patch("par_run.web.read_commands_toml", return_value=groups)

    with test_client.websocket_connect("/ws") as ws:
        _res = ws.receive_json()
        assert _res


def test_ws_full_part_fail(test_client: TestClient, mocker: MockerFixture, anyio_backend: AnyIOBackendT) -> None:  # noqa: ARG001
    command1 = Command(name="test1", cmd="echo 'Hello, World!'")
    command2 = Command(name="test2", cmd="echo 'World, Hey!' && exit 1")
    commands = OrderedDict()
    commands[command1.name] = command1
    commands[command2.name] = command2
    groups = [CommandGroup(name="test_group", cmds=commands)]

    mocker.patch("par_run.web.read_commands_toml", return_value=groups)

    with test_client.websocket_connect("/ws") as ws:
        _res = ws.receive_json()
        assert _res


async def test_get_cfg(test_client: TestClient) -> None:
    resp = test_client.get("/get-commands-config")
    assert resp.status_code == requests.codes.ok
    assert resp.json()



from collections import OrderedDict
from pathlib import Path
from typing import Any

import anyio
import pytest
import tomlkit

from par_run.executor import (
    Command,
    CommandGroup,
    CommandStatus,
    ProcessingStrategy,
    _get_optional_keys,
    _validate_mandatory_keys,
    read_commands_toml,
)
from par_run.executor import (
    internal_error_ret_code as internal_err_ret_code,
)
from py_tests.conftest import AnyIOBackendT


def test_command_incr_line_count() -> None:
    command = Command(name="test", cmd="echo 'Hello, World!'")
    assert command.num_non_empty_lines == 0

    command.incr_line_count("Hello, World!")
    assert command.num_non_empty_lines == 1

    command.incr_line_count("")
    assert command.num_non_empty_lines == 1


def test_command_append_unflushed() -> None:
    command = Command(name="test", cmd="echo 'Hello, World!'")
    assert command.unflushed == []

    command.append_unflushed("Hello, World!")
    assert command.unflushed == ["Hello, World!"]

    command.append_unflushed("")
    assert command.unflushed == ["Hello, World!", ""]


def test_command_set_running() -> None:
    command = Command(name="test", cmd="echo 'Hello, World!'")
    assert command.status == CommandStatus.NOT_STARTED

    command.set_running()
    assert command.status == CommandStatus.RUNNING  # type: ignore


class TestCommandCB:
    async def on_start(self, cmd: Command) -> None:
        assert cmd
        assert cmd.name.startswith("test")

    async def on_recv(self, cmd: Command, output: str) -> None:
        assert cmd
        assert cmd.name.startswith("test")
        assert output is not None

    async def on_term(self, cmd: Command, exit_code: int) -> None:
        assert cmd
        assert cmd.name.startswith("test")
        assert isinstance(exit_code, int)

        """Callback function for when a command receives output"""
        if cmd.status == CommandStatus.SUCCESS:
            assert cmd.status.completed()
            assert cmd.ret_code == 0
        elif cmd.status == CommandStatus.FAILURE:
            assert cmd.status.completed()
            assert cmd.ret_code != 0


def test_command_group_parallel(anyio_backend: AnyIOBackendT) -> None:  # noqa: ARG001, ANN001
    command1 = Command(name="test1", cmd="echo 'Hello, World!'", passenv=["PATH"])
    command2 = Command(name="test2", cmd="echo 'World, Hey!'", setenv={"TEST": "test"})
    commands = OrderedDict()
    commands[command1.name] = command1
    commands[command2.name] = command2
    group = CommandGroup(name="test_group", cmds=commands)
    anyio.run(group.run, ProcessingStrategy.ON_COMP, TestCommandCB())
    assert all(cmd.status.completed() for cmd in group.cmds.values())
    assert all(cmd.ret_code == 0 for cmd in group.cmds.values())
    assert all(cmd.num_non_empty_lines == 1 for cmd in group.cmds.values())
    assert all(cmd.unflushed == [] for cmd in group.cmds.values())
    assert all(cmd.status == CommandStatus.SUCCESS for cmd in group.cmds.values())
    assert all(cmd.ret_code == 0 for cmd in group.cmds.values())

    command1 = Command(name="test1", cmd="echo 'Hello, World!'")
    command2 = Command(name="test2", cmd="echo 'World, Hey!'")
    commands = OrderedDict()
    commands[command1.name] = command1
    commands[command2.name] = command2
    group = CommandGroup(name="test_group", cmds=commands)
    anyio.run(group.run, ProcessingStrategy.ON_RECV, TestCommandCB())
    assert all(cmd.status.completed() for cmd in group.cmds.values())
    assert all(cmd.ret_code == 0 for cmd in group.cmds.values())
    assert all(cmd.num_non_empty_lines == 1 for cmd in group.cmds.values())
    assert all(cmd.unflushed == [] for cmd in group.cmds.values())
    assert all(cmd.status == CommandStatus.SUCCESS for cmd in group.cmds.values())
    assert all(cmd.ret_code == 0 for cmd in group.cmds.values())


def test_command_group_serial(anyio_backend: AnyIOBackendT) -> None:  # noqa: ARG001, ANN001
    command1 = Command(name="test1", cmd="echo 'Hello, World!'", passenv=["PATH"])
    command2 = Command(name="test2", cmd="echo 'World, Hey!'", setenv={"TEST": "test"})
    commands = OrderedDict()
    commands[command1.name] = command1
    commands[command2.name] = command2
    group = CommandGroup(name="test_group", cmds=commands, serial=True)
    anyio.run(group.run, ProcessingStrategy.ON_COMP, TestCommandCB())
    assert all(cmd.status.completed() for cmd in group.cmds.values())
    assert all(cmd.ret_code == 0 for cmd in group.cmds.values())
    assert all(cmd.num_non_empty_lines == 1 for cmd in group.cmds.values())
    assert all(cmd.unflushed == [] for cmd in group.cmds.values())
    assert all(cmd.status == CommandStatus.SUCCESS for cmd in group.cmds.values())
    assert all(cmd.ret_code == 0 for cmd in group.cmds.values())

    command1 = Command(name="test1", cmd="echo 'Hello, World!'")
    command2 = Command(name="test2", cmd="echo 'World, Hey!'")
    commands = OrderedDict()
    commands[command1.name] = command1
    commands[command2.name] = command2
    group = CommandGroup(name="test_group", cmds=commands)
    anyio.run(group.run, ProcessingStrategy.ON_RECV, TestCommandCB())
    assert all(cmd.status.completed() for cmd in group.cmds.values())
    assert all(cmd.ret_code == 0 for cmd in group.cmds.values())
    assert all(cmd.num_non_empty_lines == 1 for cmd in group.cmds.values())
    assert all(cmd.unflushed == [] for cmd in group.cmds.values())
    assert all(cmd.status == CommandStatus.SUCCESS for cmd in group.cmds.values())
    assert all(cmd.ret_code == 0 for cmd in group.cmds.values())


def test_command_group_serial_part_fail(anyio_backend: AnyIOBackendT) -> None:  # noqa: ARG001, ANN001
    command1 = Command(name="test1", cmd="echo 'Hello, World!' && exit 1", passenv=["PATH"])
    command2 = Command(name="test2", cmd="echo 'World, Hey!'", setenv={"TEST": "test"})
    commands = OrderedDict()
    commands[command1.name] = command1
    commands[command2.name] = command2
    group = CommandGroup(name="test_group", cmds=commands, serial=True)

    anyio.run(group.run, ProcessingStrategy.ON_COMP, TestCommandCB())
    assert all(cmd.status.completed() for cmd in group.cmds.values())
    assert command1.ret_code == 1
    assert command2.ret_code == internal_err_ret_code
    assert command1.status == CommandStatus.FAILURE
    assert command2.status == CommandStatus.CANCELLED
    assert command1.num_non_empty_lines == 1
    assert command2.num_non_empty_lines == 0
    assert all(cmd.unflushed == [] for cmd in group.cmds.values())

    command1 = Command(name="test1", cmd="echo 'Hello, World!' && exit 1")
    command2 = Command(name="test2", cmd="echo 'World, Hey!'")
    commands = OrderedDict()
    commands[command1.name] = command1
    commands[command2.name] = command2
    group = CommandGroup(name="test_group", cmds=commands, serial=True)
    anyio.run(group.run, ProcessingStrategy.ON_RECV, TestCommandCB())
    assert all(cmd.status.completed() for cmd in group.cmds.values())
    assert command1.ret_code == 1
    assert command2.ret_code == internal_err_ret_code
    assert command1.status == CommandStatus.FAILURE
    assert command2.status == CommandStatus.CANCELLED
    assert command1.num_non_empty_lines == 1
    assert command2.num_non_empty_lines == 0
    assert all(cmd.unflushed == [] for cmd in group.cmds.values())


@pytest.fixture(
    params=[
        pytest.param(("asyncio", {"use_uvloop": True}), id="asyncio+uvloop"),
        pytest.param(("asyncio", {"use_uvloop": False}), id="asyncio"),
        pytest.param(("trio", {}), id="trio"),
    ]
)
def anyio_backend_asyncio(request: pytest.FixtureRequest) -> tuple[str, dict[str, Any]]:
    return request.param


def test_command_group_timeout_on_recv(anyio_backend_asyncio) -> None:  # noqa: ARG001, ANN001
    command1 = Command(name="test1", cmd="echo 'Hello, World!' && sleep 2 && exit 0", passenv=["PATH"])
    commands = OrderedDict()
    commands[command1.name] = command1
    group = CommandGroup(name="test_group", cmds=commands, timeout=1)
    anyio.run(group.run, ProcessingStrategy.ON_RECV, TestCommandCB())

    assert all(cmd.status.completed() for cmd in group.cmds.values())
    assert all(cmd.ret_code == internal_err_ret_code for cmd in group.cmds.values())
    assert all(cmd.num_non_empty_lines == 1 for cmd in group.cmds.values())
    assert all(cmd.unflushed == [] for cmd in group.cmds.values())
    assert all(cmd.status == CommandStatus.TIMEOUT for cmd in group.cmds.values())


def test_command_group_timeout_on_comp(anyio_backend_asyncio) -> None:  # noqa: ARG001, ANN001
    command1 = Command(name="test1", cmd="echo 'Hello, World!' && sleep 2 && exit 0", passenv=["PATH"])
    commands = OrderedDict()
    commands[command1.name] = command1
    group = CommandGroup(name="test_group", cmds=commands, timeout=1)
    anyio.run(group.run, ProcessingStrategy.ON_COMP, TestCommandCB())
    timeout_ret_code = 999
    assert all(cmd.status.completed() for cmd in group.cmds.values())
    assert all(cmd.ret_code == timeout_ret_code for cmd in group.cmds.values())
    assert all(cmd.num_non_empty_lines == 1 for cmd in group.cmds.values())
    assert all(len(cmd.unflushed) > 0 for cmd in group.cmds.values())
    assert all(cmd.status == CommandStatus.TIMEOUT for cmd in group.cmds.values())


def test_command_group_part_fail(anyio_backend: AnyIOBackendT) -> None:  # noqa: ARG001, ANN001
    command1 = Command(name="test1", cmd="echo 'Hello, World!'")
    command2 = Command(name="test2", cmd="echo 'World, Hey!'; exit 1")
    commands = OrderedDict()
    commands[command1.name] = command1
    commands[command2.name] = command2
    group = CommandGroup(name="test_group", cmds=commands)
    anyio.run(group.run, ProcessingStrategy.ON_COMP, TestCommandCB())
    assert all(cmd.status.completed() for cmd in group.cmds.values())
    assert all(cmd.ret_code in [0, 1] for cmd in group.cmds.values())
    assert all(cmd.num_non_empty_lines == 1 for cmd in group.cmds.values())
    assert all(cmd.unflushed == [] for cmd in group.cmds.values())
    assert all(cmd.status in [CommandStatus.SUCCESS, CommandStatus.FAILURE] for cmd in group.cmds.values())
    assert all(cmd.ret_code in [0, 1] for cmd in group.cmds.values())

    command1 = Command(name="test1", cmd="echo 'Hello, World!'")
    command2 = Command(name="test2", cmd="echo 'World, Hey!'; exit 1")
    commands = OrderedDict()
    commands[command1.name] = command1
    commands[command2.name] = command2
    group = CommandGroup(name="test_group", cmds=commands)
    anyio.run(group.run, ProcessingStrategy.ON_RECV, TestCommandCB())
    assert all(cmd.status.completed() for cmd in group.cmds.values())
    assert all(cmd.ret_code in [0, 1] for cmd in group.cmds.values())
    assert all(cmd.num_non_empty_lines == 1 for cmd in group.cmds.values())
    assert all(cmd.unflushed == [] for cmd in group.cmds.values())
    assert all(cmd.status in [CommandStatus.SUCCESS, CommandStatus.FAILURE] for cmd in group.cmds.values())
    assert all(cmd.ret_code in [0, 1] for cmd in group.cmds.values())


async def test_command_group_async(anyio_backend: AnyIOBackendT) -> None:  # noqa: ARG001, ANN001
    command1 = Command(name="test1", cmd="echo 'Hello, World!'")
    command2 = Command(name="test2", cmd="echo 'World, Hey!'")
    commands = OrderedDict()
    commands[command1.name] = command1
    commands[command2.name] = command2
    group = CommandGroup(name="test_group", cmds=commands)
    await group.run(ProcessingStrategy.ON_COMP, TestCommandCB())
    assert all(cmd.status.completed() for cmd in group.cmds.values())
    assert all(cmd.ret_code == 0 for cmd in group.cmds.values())
    assert all(cmd.num_non_empty_lines == 1 for cmd in group.cmds.values())
    assert all(cmd.unflushed == [] for cmd in group.cmds.values())
    assert all(cmd.status == CommandStatus.SUCCESS for cmd in group.cmds.values())
    assert all(cmd.ret_code == 0 for cmd in group.cmds.values())

    command1 = Command(name="test1", cmd="echo 'Hello, World!'")
    command2 = Command(name="test2", cmd="echo 'World, Hey!'")
    commands = OrderedDict()
    commands[command1.name] = command1
    commands[command2.name] = command2
    group = CommandGroup(name="test_group", cmds=commands)
    await group.run(ProcessingStrategy.ON_RECV, TestCommandCB())
    assert all(cmd.status.completed() for cmd in group.cmds.values())
    assert all(cmd.ret_code == 0 for cmd in group.cmds.values())
    assert all(cmd.num_non_empty_lines == 1 for cmd in group.cmds.values())
    assert all(cmd.unflushed == [] for cmd in group.cmds.values())
    assert all(cmd.status == CommandStatus.SUCCESS for cmd in group.cmds.values())
    assert all(cmd.ret_code == 0 for cmd in group.cmds.values())


async def test_command_group_async_part_fail(anyio_backend: AnyIOBackendT) -> None:  # noqa: ARG001, ANN001
    command1 = Command(name="test1", cmd="echo 'Hello, World!'")
    command2 = Command(name="test2", cmd="echo 'World, Hey!'; exit 1")
    commands = OrderedDict()
    commands[command1.name] = command1
    commands[command2.name] = command2
    group = CommandGroup(name="test_group", cmds=commands)
    await group.run(ProcessingStrategy.ON_COMP, TestCommandCB())
    assert all(cmd.status.completed() for cmd in group.cmds.values())
    assert all(cmd.ret_code in [0, 1] for cmd in group.cmds.values())
    assert all(cmd.num_non_empty_lines == 1 for cmd in group.cmds.values())
    assert all(cmd.unflushed == [] for cmd in group.cmds.values())
    assert all(cmd.status in [CommandStatus.SUCCESS, CommandStatus.FAILURE] for cmd in group.cmds.values())
    assert all(cmd.ret_code in [0, 1] for cmd in group.cmds.values())
    assert group.status == CommandStatus.FAILURE

    command1 = Command(name="test1", cmd="echo 'Hello, World!'")
    command2 = Command(name="test2", cmd="echo 'World, Hey!'; exit 1")
    commands = OrderedDict()
    commands[command1.name] = command1
    commands[command2.name] = command2
    group = CommandGroup(name="test_group", cmds=commands)
    await group.run(ProcessingStrategy.ON_RECV, TestCommandCB())
    assert all(cmd.status.completed() for cmd in group.cmds.values())
    assert all(cmd.ret_code in [0, 1] for cmd in group.cmds.values())
    assert all(cmd.num_non_empty_lines == 1 for cmd in group.cmds.values())
    assert all(cmd.unflushed == [] for cmd in group.cmds.values())
    assert all(cmd.status in [CommandStatus.SUCCESS, CommandStatus.FAILURE] for cmd in group.cmds.values())
    assert all(cmd.ret_code in [0, 1] for cmd in group.cmds.values())
    assert group.status == CommandStatus.FAILURE


def test_validate_mandatory_keys() -> None:
    data = tomlkit.table()
    data["key1"] = "value1"
    data["key2"] = "value2"
    data["key3"] = "value3"

    keys = ["key1", "key2", "key3"]
    context = "test_context"

    result = _validate_mandatory_keys(data, keys, context)
    assert result == ("value1", "value2", "value3")


def test_validate_mandatory_keys_missing_key() -> None:
    data = tomlkit.table()
    data["key1"] = "value1"
    data["key3"] = "value3"

    keys = ["key1", "key2", "key3"]
    context = "test_context"

    with pytest.raises(ValueError) as exc_info:
        _validate_mandatory_keys(data, keys, context)

    assert str(exc_info.value) == "key2 is mandatory, not found in test_context"


def test__get_optional_keys() -> None:
    data = tomlkit.table()
    data["key1"] = "value1"
    data["key2"] = "value2"
    data["key3"] = "value3"

    keys = ["key1", "key2", "key3"]
    expected_result1 = ("value1", "value2", "value3")
    assert _get_optional_keys(data, keys) == expected_result1

    keys = ["key1", "key2", "key4"]
    expected_result2 = ("value1", "value2", None)
    assert _get_optional_keys(data, keys) == expected_result2

    keys = ["key4", "key5", "key6"]
    expected_result3 = (None, None, None)
    assert _get_optional_keys(data, keys) == expected_result3

    keys = []
    expected_result4 = ()
    assert _get_optional_keys(data, keys) == expected_result4

    data = tomlkit.table()
    keys = ["key1", "key2", "key3"]
    expected_result5 = (None, None, None)
    assert _get_optional_keys(data, keys) == expected_result5


@pytest.mark.parametrize("filename", ["pyproject.toml", "commands.toml"])
def test_read_commands_toml(filename: str) -> None:
    command_groups = read_commands_toml(filename)
    assert isinstance(command_groups, list)
    assert len(command_groups) > 0
    for group in command_groups:
        assert isinstance(group, CommandGroup)
        assert group.name
        assert isinstance(group.desc, str) or group.desc is None
        assert isinstance(group.cmds, OrderedDict)
        assert isinstance(group.timeout, int)
        assert isinstance(group.cont_on_fail, bool)
        assert isinstance(group.serial, bool)

        for cmd_name, cmd in group.cmds.items():
            assert isinstance(cmd, Command)
            assert cmd.name == cmd_name
            assert cmd.cmd

    # Test case 2: Invalid commands.toml file
    with pytest.raises(FileNotFoundError):
        filename = "invalid_commands.toml"
        read_commands_toml(filename)


def test_read_commands_toml_missing_section(tmp_path: Path) -> None:
    # Create a valid TOML file without the par-run section
    toml_content = """
[BAD_HEADER]
desc = "par-run from pyproject.toml"
"""

    toml_file = tmp_path / "test_commands.toml"
    toml_file.write_text(toml_content)

    # Call the function and assert that it raises a ValueError
    with pytest.raises(ValueError):
        read_commands_toml(toml_file)


def test_read_commands_toml_pyproj(tmp_path: Path) -> None:
    # Create a valid TOML file without the par-run section
    toml_content = """
[tool.par-run]
desc = "par-run from pyproject.toml"

[[tool.par-run.groups]]
name = "Formatting"
desc = "Code formatting commands."
cont_on_fail = true # default is false
serial = false # default is false
timeout = 5

  [[tool.par-run.groups.commands]]
  name = "ruff_fmt"
  exec = "ruff format src py_tests"
  # Define an empty extras using standard table notation, or simply omit it if it's always empty

  [[tool.par-run.groups.commands]]
  name = "ruff_fix"
  exec = "ruff check --fix src py_tests"

[[tool.par-run.groups]]
name = "Formatting2"
desc = "Code formatting commands."
cont_on_fail = true # default is false
serial = true # default is false
timeout = 10

  [[tool.par-run.groups.commands]]
  name = "ruff_fmt"
  exec = "ruff format src py_tests"
  # Define an empty extras using standard table notation, or simply omit it if it's always empty

  [[tool.par-run.groups.commands]]
  name = "ruff_fix"
  exec = "ruff check --fix src py_tests"
    """

    toml_file = tmp_path / "pyproject.toml"
    toml_file.write_text(toml_content)

    num_groups = 2
    group_1_timeout = 5
    group_2_timeout = 10

    groups = read_commands_toml(toml_file)
    assert len(groups) == num_groups
    assert groups[0].name == "Formatting"
    assert groups[0].timeout == group_1_timeout
    assert groups[0].serial is False

    assert groups[1].name == "Formatting2"
    assert groups[1].timeout == group_2_timeout
    assert groups[1].serial


def test_read_commands_toml_non_pyproj(tmp_path: Path) -> None:
    # Create a valid TOML file without the par-run section
    toml_content = """
[par-run]
desc = "par-run from pyproject.toml"

[[par-run.groups]]
name = "Formatting"
desc = "Code formatting commands."
cont_on_fail = true # default is false
serial = false # default is false
timeout = 5

  [[par-run.groups.commands]]
  name = "ruff_fmt"
  exec = "ruff format src py_tests"
  # Define an empty extras using standard table notation, or simply omit it if it's always empty

  [[par-run.groups.commands]]
  name = "ruff_fix"
  exec = "ruff check --fix src py_tests"

[[par-run.groups]]
name = "Formatting2"
desc = "Code formatting commands."
cont_on_fail = true # default is false
serial = true # default is false
timeout = 10

  [[par-run.groups.commands]]
  name = "ruff_fmt"
  exec = "ruff format src py_tests"
  # Define an empty extras using standard table notation, or simply omit it if it's always empty

  [[par-run.groups.commands]]
  name = "ruff_fix"
  exec = "ruff check --fix src py_tests"
    """

    toml_file = tmp_path / "test_commands.toml"
    toml_file.write_text(toml_content)

    num_groups = 2
    group_1_timeout = 5
    group_2_timeout = 10

    groups = read_commands_toml(toml_file)
    assert len(groups) == num_groups
    assert groups[0].name == "Formatting"
    assert groups[0].timeout == group_1_timeout
    assert groups[0].serial is False

    assert groups[1].name == "Formatting2"
    assert groups[1].timeout == group_2_timeout
    assert groups[1].serial


def test_read_commands_toml_timeout_default(tmp_path: Path) -> None:
    # Create a valid TOML file without the par-run section
    toml_content = """
[par-run]
desc = "par-run from pyproject.toml"

[[par-run.groups]]
name = "Formatting"
desc = "Code formatting commands."
cont_on_fail = true # default is false
serial = false # default is false

  [[par-run.groups.commands]]
  name = "ruff_fmt"
  exec = "ruff format src py_tests"
  # Define an empty extras using standard table notation, or simply omit it if it's always empty

  [[par-run.groups.commands]]
  name = "ruff_fix"
  exec = "ruff check --fix src py_tests"
  # Define an empty extras using standard table notation, or simply omit it if it's always empty
    """

    toml_file = tmp_path / "test_commands.toml"
    toml_file.write_text(toml_content)

    groups = read_commands_toml(toml_file)
    default_timeout = 30
    assert groups[0].timeout == default_timeout



"""Web UI Module"""

from pathlib import Path

import rich
from fastapi import FastAPI, Request, WebSocket
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates

from .executor import Command, CommandGroup, CommandStatus, ProcessingStrategy, read_commands_toml

BASE_PATH = Path(__file__).resolve().parent

ws_app = FastAPI()
ws_app.mount("/static", StaticFiles(directory=str(BASE_PATH / "static")), name="static")
templates = Jinja2Templates(directory=str(BASE_PATH / "templates"))


@ws_app.get("/")
async def ws_main(request: Request) -> HTMLResponse:
    """Get the main page."""
    return templates.TemplateResponse("index.html", {"request": request})


@ws_app.get("/get-commands-config")
async def get_commands_config() -> list[CommandGroup]:
    """Get the commands configuration."""
    return read_commands_toml("commands.toml")


class WebCommandCB:
    """Websocket command callbacks."""

    def __init__(self, ws: WebSocket) -> None:
        self.ws = ws

    async def on_start(self, cmd: Command) -> None:
        rich.print(f"[blue bold]Started command {cmd.name}[/]")

    async def on_recv(self, cmd: Command, output: str) -> None:
        await self.ws.send_json({"commandName": cmd.name, "output": output})

    async def on_term(self, cmd: Command, exit_code: int) -> None:
        if cmd.status == CommandStatus.SUCCESS:
            rich.print(f"[green bold]Command {cmd.name} finished[/]")
        elif cmd.status == CommandStatus.FAILURE:
            rich.print(f"[red bold]Command {cmd.name} failed, {exit_code=:}[/]")
        await self.ws.send_json({"commandName": cmd.name, "output": {"ret_code": exit_code}})


@ws_app.websocket_route("/ws")
async def websocket_endpoint(websocket: WebSocket) -> None:
    """Websocket endpoint to run commands."""
    rich.print("Websocket connection")
    master_groups = read_commands_toml("commands.toml")
    await websocket.accept()
    rich.print("Websocket accepted")
    cb = WebCommandCB(websocket)
    rich.print("Websocket command cb created")
    for grp in master_groups:
        await grp.run(ProcessingStrategy.ON_RECV, cb)
        if grp.status != CommandStatus.SUCCESS and not grp.cont_on_fail:
            break
